import datetime

from cms.api import add_plugin
from cms.models import Placeholder
from cms.models.pluginmodel import CMSPlugin
from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from cms.plugin_rendering import ContentRenderer
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core import cache
from django.test import RequestFactory, TestCase, override_settings

from djangocms_oidc.cms_plugins import (
    OIDCDisplayDedicatedContentPlugin,
    OIDCHandoverDataPlugin,
    OIDCListIdentifiersPlugin,
    OIDCLoginPlugin,
    OIDCShowAttributeCountryPlugin,
    OIDCShowAttributePlugin,
)
from djangocms_oidc.constants import DJANGOCMS_USER_SESSION_KEY
from djangocms_oidc.models import (
    OIDCDisplayDedicatedContent,
    OIDCHandoverData,
    OIDCIdentifier,
    OIDCLogin,
    OIDCProvider,
    OIDCShowAttribute,
)


class CreateInstancesMixin:

    def _create_model(self, plugin_class, **kwargs):
        placeholder = Placeholder.objects.create(slot='test')
        model_instance = add_plugin(
            placeholder,
            plugin_class,
            'en',
            **kwargs
        )
        return model_instance

    def _create_request(self):
        request = RequestFactory().request()
        request.session = {}
        return request


class CreateProviderTestCase(CreateInstancesMixin, TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="example_id", client_secret="client_secret",
            token_endpoint="https://foo.foo/token", user_endpoint="https://foo.foo/user")


class TestOIDCHandoverDataPlugin(CreateProviderTestCase):

    def test_get_verified_as(self):
        user_info = {}
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, claims={})
        plugin_instance = model_instance.get_plugin_class_instance()
        name = plugin_instance.get_verified_as(model_instance, user_info)
        self.assertEqual(name, 'User')

    def test_get_verified_as_email(self):
        user_info = {'email': 'mail@foo.foo', 'name': 'Tester'}
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, verified_by="email",
                                            claims={})
        plugin_instance = model_instance.get_plugin_class_instance()
        name = plugin_instance.get_verified_as(model_instance, user_info)
        self.assertEqual(name, 'mail@foo.foo')

    def test_plugin_context(self):
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, claims={})
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCHandoverData)
        self.assertEqual(context['registration_consumer_info'], {
            'client_id': 'example_id', 'expires_at': None, 'consumer_type': 'MANAGED'})

    def test_plugin_context_with_user_info(self):
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, verified_by="name",
                                            claims={})
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCHandoverData)
        self.assertEqual(context['registration_consumer_info'], {
            'client_id': 'example_id', 'expires_at': None, 'consumer_type': 'MANAGED'})
        self.assertEqual(context['djangocms_oidc_user_info'], {'email': 'mail@foo.foo', 'name': 'Tester'})
        self.assertTrue(context['all_required_handovered'])
        self.assertEqual(context['djangocms_oidc_verified_as'], 'Tester')

    def test_plugin_html(self):
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, claims={})
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                <a href="/oidc-sign-up/handover-{model_id}/" class="djangocms-oidc-signup provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Data handover from Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info(self):
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, verified_by="name",
                                            claims={})
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                You are verified as <span class="djangocms-oidc-verified_as">Tester</span>.
                <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider" title="Dismiss data from Provider.">
                    Dismiss
                </a>
                <a href="/oidc-sign-up/handover-{model_id}/" class="djangocms-oidc-signup update-data provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Update data
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info_and_urls_all_required_handovered(self):
        model_instance = self._create_model(OIDCHandoverDataPlugin, provider=self.provider, verified_by="name",
                                            claims={})
        model_instance.provider.account_url = "https://provider.foo/account"
        model_instance.provider.logout_url = "https://provider.foo/logout"
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                    You are verified as
                    <span class="djangocms-oidc-verified_as">
                        <a href="https://provider.foo/account" class="djangocms-oidc-account provider">Tester</a>
                    </span>.
                    <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider"
                            title="Dismiss data from Provider.">
                        Dismiss
                    </a>
                    <a href="/oidc-sign-up/handover-{model_id}/" class="djangocms-oidc-signup update-data provider"
                            data-consumer_type="MANAGED"
                            data-client_id="example_id"
                            data-expires_at="">
                        Update data
                    </a>
                    <div>
                        You can also log out of provider
                        <a href="https://provider.foo/logout" class="djangocms-oidc-logout provider">Provider</a>
                        if you wish.
                    </div>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info_and_urls(self):
        claims = {
            'userinfo': {"email": {"essential": True}},
        }
        model_instance = self._create_model(
            OIDCHandoverDataPlugin, provider=self.provider, verified_by="name", claims=claims)
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                You are verified as <span class="djangocms-oidc-verified_as">Tester</span>.
                <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider" title="Dismiss data from Provider.">
                    Dismiss
                </a>
                <span class="all-required-not-handovered">
                    <span>Not all required data has been handovered.</span>
                    <a href="/oidc-sign-up/handover-{model_id}consent/"
                            class="djangocms-oidc-signup remaining-data provider"
                            data-consumer_type="MANAGED"
                            data-client_id="example_id"
                            data-expires_at="">
                        Handover the remaining data
                    </a>
                </span>
            </span>""".format(model_id=model_instance.pk))

    @override_settings(CMS_CACHE_PREFIX='prefix:')
    @override_settings(CACHES={'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}})
    def test_plugin_html_with_automatic_registration(self):
        provider = OIDCProvider.objects.create(name="Autoregistraion Provider", slug="autoreg-provider")
        placeholder = Placeholder.objects.create(slot='test')
        model_instance = add_plugin(placeholder, OIDCHandoverDataPlugin, 'en', provider=provider, claims={})
        request = self._create_request()
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', {
            'client_id': 42,
            'expires_at': datetime.datetime(2020, 10, 7, 14, 42, 21)
        })
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                <a href="/oidc-sign-up/handover-{model_id}/" class="djangocms-oidc-signup autoreg-provider"
                        data-consumer_type="AUTOMATIC"
                        data-client_id="42"
                        data-expires_at="2020-10-07T14:42:21">
                    Data handover from Autoregistraion Provider
                </a>
            </span>""".format(model_id=model_instance.pk))


class TestOIDCLoginPluginPlugin(CreateProviderTestCase):

    def test_plugin_context(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, claims={})
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCLogin)
        self.assertEqual(context['registration_consumer_info'], {
            'client_id': 'example_id', 'expires_at': None, 'consumer_type': 'MANAGED'})

    def test_plugin_html(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, claims={})
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Login by Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, verified_by="name", claims={})
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                You are verified as <span class="djangocms-oidc-verified_as">Tester</span>.
                <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider" title="Dismiss data from Provider.">
                    Dismiss
                </a>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup update-data provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Update data
                </a>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Login by Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info_and_urls_all_required_handovered(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, verified_by="name", claims={})
        model_instance.provider.account_url = "https://provider.foo/account"
        model_instance.provider.logout_url = "https://provider.foo/logout"
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                You are verified as
                <span class="djangocms-oidc-verified_as">
                    <a href="https://provider.foo/account" class="djangocms-oidc-account provider">Tester</a>
                </span>.
                <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider" title="Dismiss data from Provider.">
                    Dismiss
                </a>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup update-data provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Update data
                </a>
                <div>
                    You can also log out of provider
                    <a href="https://provider.foo/logout" class="djangocms-oidc-logout provider">Provider</a>
                    if you wish.
                </div>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Login by Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_with_user_info_and_urls(self):
        claims = {
            'userinfo': {"email": {"essential": True}},
        }
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, verified_by="name", claims=claims)
        model_instance.provider.account_url = "https://provider.foo/account"
        model_instance.provider.logout_url = "https://provider.foo/logout"
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                You are verified as
                <span class="djangocms-oidc-verified_as">
                    <a href="https://provider.foo/account" class="djangocms-oidc-account provider">Tester</a>
                </span>.
                <a href="/oidc-dismiss/" class="djangocms-oidc-dismiss provider" title="Dismiss data from Provider.">
                    Dismiss
                </a>
                <span class="all-required-not-handovered">
                    <span>Not all required data has been handovered.</span>
                    <a href="/oidc-sign-up/login-{model_id}consent/"
                            class="djangocms-oidc-signup remaining-data provider">
                        Handover the remaining data
                    </a>
                </span>
                <div>
                    You can also log out of provider
                    <a href="https://provider.foo/logout" class="djangocms-oidc-logout provider">Provider</a>
                    if you wish.
                </div>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup provider"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Login by Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_no_new_user(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, no_new_user=True, claims={})
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                <span class="pair-info">
                    <a href="/login/" class="djangocms-oidc-login"
                        title="To pair your account with an identity provider, sign in first.">
                        Login
                    </a>
                </span>
                <a href="/oidc-sign-up/login-{model_id}/" class="djangocms-oidc-signup provider login-required"
                        data-consumer_type="MANAGED"
                        data-client_id="example_id"
                        data-expires_at="">
                    Pair by Provider
                </a>
            </span>""".format(model_id=model_instance.pk))

    def test_plugin_html_user_is_authenticated(self):
        model_instance = self._create_model(OIDCLoginPlugin, provider=self.provider, claims={})
        request = self._create_request()
        request.user = get_user_model()()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request, 'user': request.user})
        self.assertHTMLEqual(html, """
            <span class="djangocms-oidc">
                <a href="/oidc-sign-up/login-{model_id}/"
                        class="djangocms-oidc-signup provider"
                        title="Pair my account with an identity provider">
                    Pair by Provider
                </a>
                <a href="/oidc-logout/" class="logout-from-editation">Logout</a>
            </span>""".format(model_id=model_instance.pk))


@plugin_pool.register_plugin
class TestContentPlugin(CMSPluginBase):
    model = CMSPlugin
    name = "Test Content Plugin."
    render_template = "test_content_plugin.html"


class TestOIDCDisplayDedicatedContentPlugin(CreateInstancesMixin, TestCase):

    def _create_plugin_content(self, parent_instance, **attrs):
        instance = CMSPlugin(
            parent=parent_instance,
            placeholder=parent_instance.placeholder,
            language=parent_instance.language,
            position=CMSPlugin.objects.filter(parent=parent_instance).count(),
            plugin_type=TestContentPlugin.__name__,
            **attrs
        )
        instance.save()
        if parent_instance.child_plugin_instances is None:
            parent_instance.child_plugin_instances = [instance]
        else:
            parent_instance.child_plugin_instances.append(instance)
        return instance

    def test_plugin_context_conditions_empty(self):
        model_instance = self._create_model(OIDCDisplayDedicatedContentPlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCDisplayDedicatedContent)
        self.assertFalse(context['content_permitted_to_user'])

    def test_plugin_context_no_permissions(self):
        model_instance = self._create_model(OIDCDisplayDedicatedContentPlugin, conditions='only_authenticated_user')
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.user = get_user_model()()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCDisplayDedicatedContent)
        self.assertTrue(context['content_permitted_to_user'])

    def test_plugin_html_conditions_email_verified(self):
        model_instance = self._create_model(OIDCDisplayDedicatedContentPlugin, conditions='email_verified')
        self._create_plugin_content(model_instance)
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'email_verified': True}
        request.current_page = None
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<p>Test Content Plugin.</p>")

    def test_plugin_html_conditions_uset_authenticated(self):
        model_instance = self._create_model(OIDCDisplayDedicatedContentPlugin)
        self._create_plugin_content(model_instance)
        request = self._create_request()
        request.current_page = None
        request.user = get_user_model()()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request, 'user': request.user})
        self.assertHTMLEqual(html, "<p>Test Content Plugin.</p>")

    def test_plugin_html_conditions_email_not_verified(self):
        model_instance = self._create_model(OIDCDisplayDedicatedContentPlugin, conditions='email_verified')
        self._create_plugin_content(model_instance)
        self._create_plugin_content(model_instance)
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo'}
        request.current_page = None
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<ul class='messagelist'><li class='error'>Email is not verified.</li></ul>")


class TestOIDCListIdentifiersPlugin(CreateProviderTestCase):

    def test_plugin_context_user_is_not_authenticated(self):
        model_instance = self._create_model(OIDCListIdentifiersPlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.user = AnonymousUser()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], CMSPlugin)
        self.assertIsNone(context.get('user_has_identifiers'))

    def test_plugin_context_user_is_authenticated(self):
        model_instance = self._create_model(OIDCListIdentifiersPlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.user = get_user_model()()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], CMSPlugin)
        self.assertFalse(context['user_has_identifiers'])
        self.assertQuerysetEqual(context['formset'].get_queryset().values_list('uident', flat=True), [])

    def test_plugin_context_user_with_identifier(self):
        model_instance = self._create_model(OIDCListIdentifiersPlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        User = get_user_model()
        request.user = User.objects.create(username="admin")
        OIDCIdentifier.objects.create(user=request.user, provider=self.provider, uident="1234567890")
        another_user = User.objects.create(username="other")
        OIDCIdentifier.objects.create(user=another_user, provider=self.provider, uident="42")
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], CMSPlugin)
        self.assertTrue(context['user_has_identifiers'])
        self.assertQuerysetEqual(context['formset'].get_queryset().values_list('uident', flat=True), [
            '1234567890'], transform=str)

    def test_plugin_html(self):
        model_instance = self._create_model(OIDCListIdentifiersPlugin)
        request = self._create_request()
        request.user = get_user_model()()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request, 'user': request.user})
        self.assertHTMLEqual(html, """
            <p class="djangocms-oidc-no-linked-identifiers">
                You have no identifiers associated with the provider.
            </p>""")

    def test_plugin_html_with_identifiers(self):
        model_instance = self._create_model(OIDCListIdentifiersPlugin)
        request = self._create_request()
        User = get_user_model()
        request.user = User.objects.create(username="admin")
        ident = OIDCIdentifier.objects.create(user=request.user, provider=self.provider, uident="1234567890")
        another_user = User.objects.create(username="other")
        OIDCIdentifier.objects.create(user=another_user, provider=self.provider, uident="42")
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request, 'user': request.user})
        self.assertHTMLEqual(html, """
            <form class="djangocms-oidc-delete-identifiers" method="post" action="/oidc-delete-identifiers/">
                <input type="hidden" name="form-TOTAL_FORMS" value="1" id="id_form-TOTAL_FORMS">
                <input type="hidden" name="form-INITIAL_FORMS" value="1" id="id_form-INITIAL_FORMS">
                <input type="hidden" name="form-MIN_NUM_FORMS" value="0" id="id_form-MIN_NUM_FORMS">
                <input type="hidden" name="form-MAX_NUM_FORMS" value="1000" id="id_form-MAX_NUM_FORMS">
                <ul class="user-identifiers">
                    <li>
                        <input type="hidden" name="form-0-cmsplugin_ptr" value="{id}" id="id_form-0-cmsplugin_ptr">
                        <input type="checkbox" name="form-0-DELETE" id="id_form-0-DELETE">
                        <span class="user-identity">1234567890</span>
                    </li>
                </ul>
                <input type="submit" name="djangocms_oidc_delete_identifier" value="Delete selected">
            </form>""".format(id=ident.pk))


class TestOIDCShowAttributePlugin(CreateInstancesMixin, TestCase):

    def test_plugin_context(self):
        model_instance = self._create_model(OIDCShowAttributePlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCShowAttribute)
        self.assertIsNone(context['djangocms_oidc_verified_as'])

    def test_plugin_context_with_default(self):
        model_instance = self._create_model(OIDCShowAttributePlugin, default_value="Default")
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCShowAttribute)
        self.assertEqual(context['djangocms_oidc_verified_as'], 'Default')

    def test_plugin_context_from_session(self):
        model_instance = self._create_model(OIDCShowAttributePlugin, default_value="Default", verified_by="email")
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCShowAttribute)
        self.assertEqual(context['djangocms_oidc_verified_as'], 'mail@foo.foo')

    def test_plugin_html(self):
        model_instance = self._create_model(OIDCShowAttributePlugin)
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<span></span>")

    def test_plugin_html_with_default(self):
        model_instance = self._create_model(OIDCShowAttributePlugin, default_value="Default")
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<span>Default</span>")

    def test_plugin_html_from_session(self):
        model_instance = self._create_model(OIDCShowAttributePlugin, default_value="Default", verified_by="email")
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'email': 'mail@foo.foo', 'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<span>mail@foo.foo</span>")


class TestOIDCShowAttributeCountryPlugin(CreateInstancesMixin, TestCase):

    def test_plugin_context(self):
        model_instance = self._create_model(OIDCShowAttributeCountryPlugin)
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCShowAttribute)
        self.assertIsNone(context['djangocms_oidc_verified_as'])

    def test_plugin_context_from_session(self):
        model_instance = self._create_model(OIDCShowAttributeCountryPlugin, verified_by="country")
        plugin_instance = model_instance.get_plugin_class_instance()
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'country': 'CZ', 'name': 'Tester'}
        context = plugin_instance.render({'request': request}, model_instance, None)
        self.assertIsInstance(context['instance'], OIDCShowAttribute)
        self.assertEqual(context['djangocms_oidc_verified_as'], 'CZ')

    def test_plugin_html(self):
        model_instance = self._create_model(OIDCShowAttributeCountryPlugin)
        request = self._create_request()
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertEqual(html.strip(), "")

    def test_plugin_html_from_session(self):
        model_instance = self._create_model(OIDCShowAttributeCountryPlugin, verified_by="country")
        request = self._create_request()
        request.session[DJANGOCMS_USER_SESSION_KEY] = {'country': 'CZ', 'name': 'Tester'}
        renderer = ContentRenderer(request=RequestFactory())
        html = renderer.render_plugin(model_instance, {'request': request})
        self.assertHTMLEqual(html, "<span>Czechia</span>")



from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase

from djangocms_oidc import helpers
from djangocms_oidc.constants import DJANGOCMS_PLUGIN_SESSION_KEY, DJANGOCMS_USER_SESSION_KEY
from djangocms_oidc.models import OIDCHandoverData, OIDCIdentifier, OIDCProvider


class TestHelpers(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="example_id", client_secret="client_secret",
            token_endpoint="https://foo.foo/token", user_endpoint="https://foo.foo/user")
        cls.plugin = OIDCHandoverData.objects.create(provider=cls.provider, claims={})
        cls.user = get_user_model().objects.create(username="admin")

    def test_set_consumer(self):
        request = RequestFactory().request()
        request.session = {}
        helpers.set_consumer(request, self.plugin)
        self.assertEqual(request.session, {DJANGOCMS_PLUGIN_SESSION_KEY: ('handover', self.plugin.pk)})

    def test_get_consumer_none(self):
        request = RequestFactory().request()
        request.session = {}
        self.assertIsNone(helpers.get_consumer(request))

    def test_get_consumer_object_does_not_exists(self):
        request = RequestFactory().request()
        request.session = {DJANGOCMS_PLUGIN_SESSION_KEY: ('handover', 42)}
        self.assertIsNone(helpers.get_consumer(request))

    def test_get_consumer(self):
        request = RequestFactory().request()
        request.session = {DJANGOCMS_PLUGIN_SESSION_KEY: ('handover', self.plugin.pk)}
        plugin = helpers.get_consumer(request)
        self.assertEqual(self.plugin, plugin)

    def test_load_consumer_unknown_type(self):
        plugin = helpers.load_consumer('foo', self.plugin.pk)
        self.assertIsNone(plugin)

    def test_load_consumer_object_does_not_exists(self):
        plugin = helpers.load_consumer('handover', 42)
        self.assertIsNone(plugin)

    def test_load_consumer(self):
        plugin = helpers.load_consumer('handover', self.plugin.pk)
        self.assertEqual(self.plugin, plugin)

    def test_get_user_identifiers_formset_empty(self):
        formset = helpers.get_user_identifiers_formset(self.user)
        self.assertFalse(formset.is_bound)
        self.assertEqual(formset.get_queryset().count(), 0)

    def test_get_user_identifiers_formset(self):
        OIDCIdentifier.objects.create(user=self.user, provider=self.provider, uident="1234567890")
        formset = helpers.get_user_identifiers_formset(self.user)
        self.assertFalse(formset.is_bound)
        self.assertQuerysetEqual(formset.get_queryset().values_list('uident', flat=True), [
            '1234567890'], transform=str)

    def test_check_required_handovered(self):
        self.plugin.claims = {
            "userinfo": {
                "openid2_id": {"essential": True},
                "email": {"essential": True},
            }
        }
        user_info = {"openid2_id": "ID", "email": "mail@foo.foo", "name": "Name"}
        self.assertTrue(helpers.check_required_handovered(self.plugin, user_info))

    def test_check_required_handovered_failsed(self):
        self.plugin.claims = {
            "userinfo": {
                "openid2_id": {"essential": True},
                "email": {"essential": True},
            }
        }
        user_info = {"openid2_id": "ID", "name": "Name"}
        self.assertFalse(helpers.check_required_handovered(self.plugin, user_info))

    def test_get_verified_as_default(self):
        name = helpers.get_verified_as(None, None, "Default")
        self.assertEqual(name, "Default")

    def test_get_verified_as_empty_no_verified_by(self):
        user_info = {"name": "Arnold Rimmer", "given_name": "Dave", "family_name": "Lister"}
        name = helpers.get_verified_as("", user_info, "Default")
        self.assertEqual(name, "Default")

    def test_get_verified_as_name(self):
        verified_by = "name given_name+family_name email"
        user_info = {"name": "Arnold Rimmer", "given_name": "Dave", "family_name": "Lister"}
        name = helpers.get_verified_as(verified_by, user_info, "Default")
        self.assertEqual(name, "Arnold Rimmer")

    def test_get_verified_as_given_and_family_name(self):
        verified_by = "given_name+family_name name email"
        user_info = {"name": "Arnold Rimmer", "given_name": "Dave", "family_name": "Lister"}
        name = helpers.get_verified_as(verified_by, user_info, "Default")
        self.assertEqual(name, "Dave Lister")

    def test_get_verified_as_email(self):
        verified_by = "given_name+family_name name email"
        user_info = {"email": "mail@foo.foo"}
        name = helpers.get_verified_as(verified_by, user_info, "Default")
        self.assertEqual(name, "mail@foo.foo")

    def test_get_user_info_is_none(self):
        request = RequestFactory().request()
        request.session = {}
        self.assertIsNone(helpers.get_user_info(request))

    def test_get_user_info(self):
        request = RequestFactory().request()
        request.session = {DJANGOCMS_USER_SESSION_KEY: 'ok'}
        self.assertEqual(helpers.get_user_info(request), 'ok')

    def test_clear_user_info_is_empty(self):
        request = RequestFactory().request()
        request.session = {}
        helpers.clear_user_info(request)
        self.assertEqual(request.session, {})

    def test_clear_user_info(self):
        request = RequestFactory().request()
        request.session = {DJANGOCMS_USER_SESSION_KEY: 'ok'}
        helpers.clear_user_info(request)
        self.assertEqual(request.session, {})



import re

from django.core.exceptions import ObjectDoesNotExist
from django.forms import modelformset_factory
from django.utils.encoding import force_str

from .constants import DJANGOCMS_PLUGIN_SESSION_KEY, DJANGOCMS_USER_SESSION_KEY
from .models import CONSUMER_CLASS, OIDCIdentifier


def set_consumer(request, instance):
    """Set consumer into session."""
    request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (instance.consumer_type, instance.pk)


def get_consumer(request):
    """Get consumer from session."""
    plugin_type_and_id = request.session.get(DJANGOCMS_PLUGIN_SESSION_KEY)
    if plugin_type_and_id is not None:
        consumer_type, instance_id = plugin_type_and_id
        try:
            return CONSUMER_CLASS[consumer_type].objects.get(pk=instance_id)
        except ObjectDoesNotExist:
            pass
    return None


def load_consumer(consumer_type, instance_id):
    """Load consumer according of type and ID."""
    if consumer_type in CONSUMER_CLASS:
        try:
            return CONSUMER_CLASS[consumer_type].objects.get(pk=instance_id)
        except ObjectDoesNotExist:
            pass
    return None


def get_user_identifiers_formset(user, post=None, files=None):
    """Get OIDCIdentifier formset."""
    IdentifierFormSet = modelformset_factory(OIDCIdentifier, fields=(), extra=0, can_delete=True)
    return IdentifierFormSet(post, files, queryset=OIDCIdentifier.objects.filter(user=user))


def check_required_handovered(consumer, user_info):
    """Check that all required data has been handovered."""
    for key, data in consumer.claims.get('userinfo', {}).items():
        if data['essential'] and key not in user_info:
            return False
    return True


def get_verified_as(verified_by, user_info, default):
    """Collect values from user_info and attributes definitions."""
    name = default
    if verified_by is not None:
        names = []
        for item in re.split(r'\s+', verified_by):
            for key in item.split('+'):
                value = user_info.get(key)
                if value:
                    names.append(force_str(value))
            if names:
                name = " ".join(names)
                break
    return name


def get_user_info(request):
    """Get user info stored in session."""
    return request.session.get(DJANGOCMS_USER_SESSION_KEY)


def clear_user_info(request):
    """Delete user info stored in session."""
    request.session.pop(DJANGOCMS_USER_SESSION_KEY, None)


def request_is_ajax(request):
    """Check that request is an Ajax type."""
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'



import datetime
import json
import re
import time
from unittest.mock import MagicMock, patch
from urllib.parse import parse_qs

from cms.api import create_page
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.signals import user_logged_out
from django.contrib.messages import get_messages
from django.contrib.messages.storage.fallback import FallbackStorage
from django.core.cache import cache
from django.dispatch import receiver
from django.test import Client, RequestFactory, TestCase, override_settings
from django.test.client import ClientHandler
from freezegun import freeze_time

from djangocms_oidc.constants import DJANGOCMS_PLUGIN_SESSION_KEY
from djangocms_oidc.middleware import OIDCSessionRefresh
from djangocms_oidc.models import OIDCHandoverData, OIDCProvider

User = get_user_model()


class OIDCSessionRefreshTokenMiddlewareTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="foo", client_secret="secret",
            token_endpoint="https://foo.foo/token", user_endpoint="https://foo.foo/user",
            authorization_endpoint='http://example.com/authorize')
        cls.plugin = OIDCHandoverData.objects.create(provider=cls.provider, claims={})

    def setUp(self):
        self.factory = RequestFactory()
        self.middleware = OIDCSessionRefresh(MagicMock)
        self.user = User.objects.create_user('example_username')

    def test_anonymous(self):
        request = self.factory.get('/foo')
        request.session = {}
        request.user = AnonymousUser()
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    def test_is_oidc_path(self):
        request = self.factory.get('/oidc/callback/')
        request.user = AnonymousUser()
        request.session = {}
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    def test_is_POST(self):
        request = self.factory.post('/foo')
        request.user = AnonymousUser()
        request.session = {}
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_is_ajax(self, mock_middleware_random):
        mock_middleware_random.return_value = 'examplestring'

        request = self.factory.get(
            '/foo',
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        request.session = {DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)}
        request.user = self.user

        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)
        # The URL to go to is available both as a header and as a key
        # in the JSON response.
        self.assertTrue(response['refresh_url'])
        url, qs = response['refresh_url'].split('?')
        self.assertEqual(url, 'http://example.com/authorize')
        expected_query = {
            'response_type': ['code'],
            'redirect_uri': ['http://testserver/callback/'],
            'client_id': ['foo'],
            'nonce': ['examplestring'],
            'prompt': ['none'],
            'scope': ['openid email'],
            'state': ['examplestring'],
        }
        self.assertEqual(expected_query, parse_qs(qs))
        json_payload = json.loads(response.content.decode('utf-8'))
        self.assertEqual(json_payload['refresh_url'], response['refresh_url'])

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120, OIDC_USE_NONCE=False)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_is_ajax_oidc_not_use_nonce_and_redirect_page(self, mock_middleware_random):
        mock_middleware_random.return_value = 'examplestring'

        request = self.factory.get(
            '/foo',
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        cms_page = create_page('test', 'test_content_plugin.html', 'en', slug="test")
        plugin = OIDCHandoverData.objects.create(provider=self.provider, redirect_page=cms_page, claims={})
        request.session = {DJANGOCMS_PLUGIN_SESSION_KEY: (plugin.consumer_type, plugin.pk)}
        request.user = self.user

        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)
        # The URL to go to is available both as a header and as a key
        # in the JSON response.
        self.assertTrue(response['refresh_url'])
        url, qs = response['refresh_url'].split('?')
        self.assertEqual(url, 'http://example.com/authorize')
        expected_query = {
            'response_type': ['code'],
            'redirect_uri': ['http://testserver/callback/'],
            'client_id': ['foo'],
            'prompt': ['none'],
            'scope': ['openid email'],
            'state': ['examplestring'],
        }
        self.assertEqual(expected_query, parse_qs(qs))
        json_payload = json.loads(response.content.decode('utf-8'))
        self.assertEqual(json_payload['refresh_url'], response['refresh_url'])

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_no_oidc_token_expiration_forces_renewal(self, mock_middleware_random):
        mock_middleware_random.return_value = 'examplestring'

        request = self.factory.get('/foo')
        request.user = self.user
        request.session = {DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)}

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split('?')
        self.assertEqual(url, 'http://example.com/authorize')
        expected_query = {
            'response_type': ['code'],
            'redirect_uri': ['http://testserver/callback/'],
            'client_id': ['foo'],
            'nonce': ['examplestring'],
            'prompt': ['none'],
            'scope': ['openid email'],
            'state': ['examplestring'],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_expired_token_forces_renewal(self, mock_middleware_random):
        mock_middleware_random.return_value = 'examplestring'

        request = self.factory.get('/foo')
        request.user = self.user
        request.session = {
            'oidc_id_token_expiration': time.time() - 10,
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk),
        }

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split('?')
        self.assertEqual(url, 'http://example.com/authorize')
        expected_query = {
            'response_type': ['code'],
            'redirect_uri': ['http://testserver/callback/'],
            'client_id': ['foo'],
            'nonce': ['examplestring'],
            'prompt': ['none'],
            'scope': ['openid email'],
            'state': ['examplestring'],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @freeze_time('2020-10-13 14:52:42')
    def test_expiration(self):
        request = self.factory.get('/foo')
        request.user = self.user
        request.session = {
            'oidc_id_token_expiration': datetime.datetime(2020, 10, 13, 14, 53, 28).timestamp()
        }
        response = self.middleware.process_request(request)
        self.assertIsNone(response)

    def _get_messages(self, request):
        return [(msg.level, msg.message) for msg in get_messages(request)]

    def test_consumet_is_none(self):
        request = self.factory.get('/foo')
        request.user = self.user
        request.session = {}
        request._messages = FallbackStorage(request)
        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['location'], '/')
        self.assertEqual(self._get_messages(request), [
            (messages.ERROR, 'OIDC Consumer activation falied. Please try again.')
        ])


def override_middleware(fun):
    classes = [
        'django.contrib.sessions.middleware.SessionMiddleware',
        'djangocms_oidc.middleware.OIDCSessionRefresh',
    ]
    return override_settings(MIDDLEWARE=classes)(fun)


class UserifiedClientHandler(ClientHandler):
    """Enhances ClientHandler to "work" with users properly"""

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user')
        super().__init__(*args, **kwargs)

    def get_response(self, req):
        req.user = self.user
        return super().get_response(req)


class ClientWithUser(Client):
    """Enhances Client to "work" with users properly"""

    def __init__(self, enforce_csrf_checks=False, **defaults):
        # Start off with the AnonymousUser
        self.user = AnonymousUser()
        # Get this because we need to create a new UserifiedClientHandler later
        self.enforce_csrf_checks = enforce_csrf_checks
        super().__init__(**defaults)
        # Stomp on the ClientHandler with one that correctly makes request.user
        # the AnonymousUser
        self.handler = UserifiedClientHandler(enforce_csrf_checks, user=self.user)

    def login(self, **credentials):
        from django.contrib.auth import authenticate

        # Try to authenticate and throw an exception if that fails; also, this gets
        # the user instance that was authenticated with
        user = authenticate(**credentials)
        if not user:
            # Client lets you fail authentication without providing any helpful
            # messages; we throw an exception because silent failure is
            # unhelpful
            raise Exception('Unable to authenticate with %r' % sorted(credentials.items()))

        ret = super().login(**credentials)
        if not ret:
            raise Exception('Login failed')

        # Stash the user object it used and rebuild the UserifiedClientHandler
        self.user = user
        self.handler = UserifiedClientHandler(self.enforce_csrf_checks, user=self.user)
        return ret


@override_middleware
class MiddlewareTestCase(TestCase):
    """These tests test the middleware as part of the request/response cycle"""

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="foo", client_secret="secret",
            token_endpoint="https://foo.foo/token", user_endpoint="https://foo.foo/user",
            authorization_endpoint='http://example.com/authorize')
        cls.plugin = OIDCHandoverData.objects.create(provider=cls.provider, claims={})

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(username='example_username', password='password')
        cache.clear()

    @override_settings(OIDC_EXEMPT_URLS=['mdo_fake_view'])
    def test_get_exempt_urls_setting_view_name(self):
        middleware = OIDCSessionRefresh(MagicMock)
        self.assertEqual(
            sorted(list(middleware.exempt_urls)),
            ['/authenticate/', '/callback/', '/logout/', '/mdo_fake_view/']
        )

    @override_settings(OIDC_EXEMPT_URLS=['/foo/'])
    def test_get_exempt_urls_setting_url_path(self):
        middleware = OIDCSessionRefresh(MagicMock)
        self.assertEqual(
            sorted(list(middleware.exempt_urls)),
            ['/authenticate/', '/callback/', '/foo/', '/logout/']
        )

    def test_is_refreshable_url(self):
        request = self.factory.get('/mdo_fake_view/')
        request.user = self.user
        request.session = dict()
        middleware = OIDCSessionRefresh(MagicMock)
        assert middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=['mdo_fake_view'])
    def test_is_not_refreshable_url_exempt_view_name(self):
        request = self.factory.get('/mdo_fake_view/')
        request.user = self.user
        request.session = dict()
        middleware = OIDCSessionRefresh(MagicMock)
        assert not middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=['/mdo_fake_view/'])
    def test_is_not_refreshable_url_exempt_path(self):
        request = self.factory.get('/mdo_fake_view/')
        request.user = self.user
        request.session = dict()
        middleware = OIDCSessionRefresh(MagicMock)
        assert not middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=[re.compile(r'^/mdo_.*_view/$')])
    def test_is_not_refreshable_url_exempt_pattern(self):
        request = self.factory.get('/mdo_fake_view/')
        request.user = self.user
        request.session = dict()
        middleware = OIDCSessionRefresh(MagicMock)
        assert not middleware.is_refreshable_url(request)

    def test_anonymous(self):
        client = ClientWithUser()
        resp = client.get('/mdo_fake_view/')
        self.assertEqual(resp.status_code, 200)

    @patch("django.contrib.auth.authenticate")
    def test_anonymous_user_not_found(self, mock_authenticate):
        mock_authenticate.return_value = None
        client = ClientWithUser()
        message = "Unable to authenticate with [('password', 'password'), ('username', 'example_username')]"
        with self.assertRaisesMessage(Exception, message):
            client.login(username=self.user.username, password='password')
        mock_authenticate.assert_called_with(username='example_username', password='password')

    @patch("django.test.Client.login")
    def test_anonymous_user_login_failed(self, mock_login):
        mock_login.return_value = None
        client = ClientWithUser()
        with self.assertRaisesMessage(Exception, 'Login failed'):
            client.login(username=self.user.username, password='password')
        mock_login.assert_called_with(username='example_username', password='password')

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    def test_authenticated_user(self):
        client = ClientWithUser()
        client.login(username=self.user.username, password='password')

        # Set the expiration to some time in the future so this user is valid
        session = client.session
        session['oidc_id_token_expiration'] = time.time() + 100
        session.save()

        resp = client.get('/mdo_fake_view/')
        self.assertEqual(resp.status_code, 200)

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_expired_token_redirects_to_sso(self, mock_middleware_random):
        mock_middleware_random.return_value = 'examplestring'

        client = ClientWithUser()
        client.login(username=self.user.username, password='password')

        # Set expiration to some time in the past
        session = client.session
        session['oidc_id_token_expiration'] = time.time() - 100
        session['_auth_user_backend'] = 'mozilla_django_oidc.auth.OIDCAuthenticationBackend'
        session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        session.save()

        resp = client.get('/mdo_fake_view/')
        self.assertEqual(resp.status_code, 302)

        url, qs = resp.url.split('?')
        self.assertEqual(url, 'http://example.com/authorize')
        expected_query = {
            'response_type': ['code'],
            'redirect_uri': ['http://testserver/callback/'],
            'client_id': ['foo'],
            'nonce': ['examplestring'],
            'prompt': ['none'],
            'scope': ['openid email'],
            'state': ['examplestring'],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch('djangocms_oidc.middleware.get_random_string')
    def test_refresh_fails_for_already_signed_in_user(self, mock_random_string):
        mock_random_string.return_value = 'examplestring'

        # Mutable to log which users get logged out.
        logged_out_users = []

        # Register a signal on 'user_logged_out' so we can
        # update 'logged_out_users'.
        @receiver(user_logged_out)
        def logged_out(sender, user=None, **kwargs):
            logged_out_users.append(user)

        client = ClientWithUser()
        # First confirm that the home page is a public page.
        resp = client.get('/')
        # At least security doesn't kick you out.
        self.assertEqual(resp.status_code, 200)
        # Also check that this page doesn't force you to redirect
        # to authenticate.
        resp = client.get('/mdo_fake_view/')
        self.assertEqual(resp.status_code, 200)
        client.login(username=self.user.username, password='password')

        # Set expiration to some time in the past
        session = client.session
        session['oidc_id_token_expiration'] = time.time() - 100
        session['_auth_user_backend'] = 'mozilla_django_oidc.auth.OIDCAuthenticationBackend'
        session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        session.save()

        # Confirm that now you're forced to authenticate again.
        resp = client.get('/mdo_fake_view/')
        self.assertEqual(resp.status_code, 302)
        self.assertTrue(
            'http://example.com/authorize' in resp.url and
            'prompt=none' in resp.url
        )
        # Now suppose the user goes there and something goes wrong.
        # For example, the user might have become "blocked" or the 2FA
        # verficiation has expired and needs to be done again.
        resp = client.get('/callback/', {
            'error': 'login_required',
            'error_description': 'Multifactor authentication required',
        })
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(resp.url, '/')

        # Since the user in 'client' doesn't change, we have to use other
        # queues to assert that the user got logged out properly.

        # The session gets flushed when you get signed out.
        # This is the only decent way to know the user lost all
        # request.session and
        self.assertTrue(not client.session.items())

        # The signal we registered should have fired for this user.
        self.assertEqual(client.user, logged_out_users[0])
